/* FILE: oxscmd.cc                 -*-Mode: c++-*-
 *
 * Tcl interface to C/C++ code in oxs.
 *
 * Each routine in the file is registered with the Tcl interpreter
 * by OxsRegisterInterfaceCommands() at the bottom of this file.
 * OxsRegisterInterfaceCommands() is called by Oxs_Init().
 * (Note: OxsRegisterInterfaceCommands() is not in any header file
 * because it should only be called by Oxs_Init().)
 *
 */

#include <string>
#include <vector>

#include "director.h"
#include "driver.h"  // For Oxs_Driver::RunEvent structure
#include "ext.h"
#include "oc.h"
#include "output.h"
#include "energy.h"	// Needed to make MSVC++ 5 happy

OC_USE_STRING;

/* End includes */     

/*
 *----------------------------------------------------------------------
 *
 * Interface commands.  These are accessed through the general purpose
 * OxsCmdsSwitchboard command.  The return value should be a string that
 * will be appended to *interp's result.  All errors should be handled
 * by throwing exceptions.  Normal return (i.e., not via an exception)
 * is interpreted as meaning sucess, and a TCL_OK return is passed
 * back to *interp.
 *  For historical reasons, some commands place their output into
 * *interp's result by themselves.  In this case the interface command
 * should return an empty string.
 *
 *---------------------------------------------------------------------- */
typedef String Oxs_CmdProc(Oxs_Director* director,Tcl_Interp *interp,
			   int argc,CONST84 char** argv);
Oxs_CmdProc Oxs_SetRestartFlag;
Oxs_CmdProc Oxs_SetRestartCrcCheck;
Oxs_CmdProc Oxs_GetMifCrc;
Oxs_CmdProc Oxs_GetMifParameters;
Oxs_CmdProc Oxs_ProbInit;
Oxs_CmdProc Oxs_ProbReset;
Oxs_CmdProc Oxs_Run;
Oxs_CmdProc Oxs_DriverProperties;
Oxs_CmdProc Oxs_ListRegisteredExtClasses;
Oxs_CmdProc Oxs_ListEnergyObjects;
Oxs_CmdProc Oxs_ListExtObjects;
Oxs_CmdProc Oxs_ListOutputObjects;
Oxs_CmdProc Oxs_OutputGet;
Oxs_CmdProc Oxs_OutputNames;
Oxs_CmdProc Oxs_SetStage;
Oxs_CmdProc Oxs_GetStage;
Oxs_CmdProc Oxs_GetRunStateCounts;
Oxs_CmdProc Oxs_GetMif;

Tcl_CmdProc OxsCmdsSwitchboard;
Tcl_CmdProc Oxs_ProbRelease;

/*
 *----------------------------------------------------------------------
 *
 * Client data struct used by Tcl_CmdProc's in oxscmds.cc
 *
 *----------------------------------------------------------------------
 */
struct OxsCmdsClientData {
public:
  Oxs_Director* director;
  Oxs_CmdProc* cmd;
  String function_name; // Used to provide context information to
  /// user when generating error messages.
  OxsCmdsClientData(Oxs_Director* d,Oxs_CmdProc* c,const String& n)
    : director(d), cmd(c), function_name(n) {}
};

/*
 *----------------------------------------------------------------------
 *
 * OxsCmdsTclException --
 *   Special exception class used by Oxs_CmdProc routines.  It indicates
 *   that the error message and other Tcl error handling info has already
 *   been set into the Tcl interpreter.  The handling routine (i.e.,
 *   OxsCmdsSwitchboard) that catches this exception should simply return
 *   control to Tcl with the specified error code (usually TCL_ERROR).
 *   This is used, for example, by header code in each Oxs_CmdProc
 *   that validates the argument count.
 *
 *----------------------------------------------------------------------
 */

class OxsCmdsProcTclException {
private:
  // operator= not possible because of const data member.
  OxsCmdsProcTclException& operator=(const OxsCmdsProcTclException&);
public:
  const int error_code;
  OxsCmdsProcTclException(int code) : error_code(code) {}
};

/*
 *----------------------------------------------------------------------
 *
 * General command cleanup routine.  Deletes clientData structure.
 *  Note: This routine is called when a command is deleted from
 *        the interpreter.  It is not called when the interpreter
 *        is destroyed as part of Oxs shutdown.  So, in practice,
 *        this routine is never called.
 *
 *----------------------------------------------------------------------
 */
Tcl_CmdDeleteProc OxsCmdsCleanup;
void OxsCmdsCleanup(ClientData clientData)
{
  OxsCmdsClientData* cd = static_cast<OxsCmdsClientData*>(clientData);
  delete cd;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_SetErrorCode --
 *      Set the error code of a Tcl interp.
 *      (Might move this to package Oc if others can use it)
 *
 *----------------------------------------------------------------------
 */

void Oxs_SetErrorCode(Tcl_Interp *interp, const char* value)
{
#if TCL_MAJOR_VERSION < 8
  Tcl_SetErrorCode(interp, value, NULL);
#else
  Tcl_SetObjErrorCode(interp, Tcl_NewStringObj(value, -1));
#endif
}

/*
 *----------------------------------------------------------------------
 *
 * General purpose switchboard command.  This routine interfaces
 * with the main Tcl interpreter, routing calls based on clientData.
 * This allows error handling to be managed at a single point.
 *
 *----------------------------------------------------------------------
 */
int OxsCmdsSwitchboard(ClientData cd,Tcl_Interp *interp,
		       int argc,CONST84 char** argv)
{
  static String result_message;
  static String error_message,error_info,error_code;
  int result_code = TCL_ERROR;
  OxsCmdsClientData* ocd = static_cast<OxsCmdsClientData*>(cd);
  try {
    result_message.erase();
    result_message = ocd->cmd(ocd->director,interp,argc,argv);
    result_code = TCL_OK;
  } catch (OxsCmdsProcTclException& err) {
    const char* ei = Tcl_GetVar(interp,Oc_AutoBuf("errorInfo"),
                                TCL_GLOBAL_ONLY);
    if(ei!=NULL) {
      error_info = String(ei);
    } else {
      error_info.erase();
    }
    const char* ec = Tcl_GetVar(interp,Oc_AutoBuf("errorCode"),
                                TCL_GLOBAL_ONLY);
    if(ec!=NULL) {
      error_code = String(ec);
    } else { 
      error_code.erase();
    }
    // In the case of this exception, the error message has been
    // set into *interp's result.  Pull it out and clear the result,
    // both for consistency with the other catch blocks, but also
    // because the Tcl_AddErrorInfo will insert the intero result
    // into the error info.
    error_message = Tcl_GetStringResult(interp);
    Tcl_ResetResult(interp); 
    result_code = err.error_code;
  } catch (Oxs_Ext::Error& err) {
    error_message =
      String("Error thrown from inside ")
      + ocd->function_name
      + String(" --- ")
      + String(err);
    error_info.erase();
    error_code.erase();
    result_code = TCL_ERROR; // Safety
  } catch (Oxs_TclError& err) {
    error_message =
      String("Error thrown from inside ")
      + ocd->function_name
      + String(" --- ")
      + String(err.MessageType())
      + String(" --- ")
      + err.MessageText();
    error_info = String(err.ErrorInfo());
    error_code = String(err.ErrorCode());
    result_code = TCL_ERROR; // Safety
  } catch (Oc_Exception& err) {
    error_message =
      String("Error thrown from inside ")
      + ocd->function_name
      + String(" --- ")
      + String(err.MessageType())
      + String(" --- ")
      + err.MessageText();
    error_info.erase();
    error_code.erase();
    result_code = TCL_ERROR; // Safety
  } catch (const char* err) {
    // Some routines in the pkg area throw bare char*'s or
    // strings.  These should be changed to Oc_Exceptions.
    error_message =
      String("Error thrown from inside"
             " unspecified OOMMF library module --- ")
      + String(err);
    error_info.erase();
    error_code.erase();
    result_code = TCL_ERROR; // Safety
  } catch (String& err) {
    error_message =
      String("Error thrown from inside"
             " unspecified OOMMF library module --- ")
      + err;
    error_info.erase();
    error_code.erase();
    result_code = TCL_ERROR; // Safety
  } catch (EXCEPTION& err) {
    error_message =
      String("Library error thrown from inside ")
      + ocd->function_name
      + String(" --- ")
      + String(err.what());
    error_info.erase();
    error_code.erase();
    result_code = TCL_ERROR; // Safety
  } catch (...) {
    error_message =
      String("Unrecognized error thrown from inside ")
      + ocd->function_name + String(".");
    error_info.erase();
    error_code.erase();
    result_code = TCL_ERROR; // Safety
  }

  if(result_code == TCL_OK) {
    // Note: Some commands may have already placed their output
    // into interp's result.  This is typically signified by
    // an empty result_message, although some commands just
    // have no output.
    if(result_message.length()>0) {
      Tcl_AppendResult(interp,result_message.c_str(),(char *)NULL);
    }
  } else { // ERROR
    // Reentrancy protection
    String tmp_error_info = error_info;       error_info.erase();
    String tmp_error_code = error_code;       error_code.erase();
    String tmp_error_message = error_message; error_message.erase();
    Oxs_TclInterpState tcl_state(interp);
    try {
      ocd->director->SetErrorStatus(1);
      Tcl_Eval(interp,Oc_AutoBuf("ReleaseProblem"));
    } catch (...) {}
    tcl_state.Restore();
    if(!tmp_error_info.empty()) {
      Tcl_AddErrorInfo(interp,tmp_error_info.c_str());
    }
    if(!tmp_error_code.empty()) {
      Oxs_SetErrorCode(interp, tmp_error_code.c_str());
      /// Use Obj interface, if available.
    }
    if(!tmp_error_message.empty()) {
      Tcl_AppendResult(interp,tmp_error_message.c_str(),NULL);
    }
  }

  return result_code;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_SetRestartFlag --
 *      Interactively set the restart flag.
 *
 * Results:
 *      Returns the previous value.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_SetRestartFlag(Oxs_Director* director,Tcl_Interp *interp,
			  int argc,CONST84 char** argv)
{
  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " restart_flag\"", (char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int newflag;
  if (Tcl_GetInt(interp, argv[1], &newflag) != TCL_OK) {
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int oldflag = director->SetRestartFlag(newflag);
  char buf[64];
  Oc_Snprintf(buf,sizeof(buf),"%d",oldflag);
  return String(buf);
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_SetRestartCrcCheck --
 *      Interactively set the restart crc check flag.
 *
 * Results:
 *      Returns the previous value.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

String Oxs_SetRestartCrcCheck(Oxs_Director* director,Tcl_Interp *interp,
			   int argc,CONST84 char** argv)
{
  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " check_flag\"", (char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int newflag;
  if (Tcl_GetInt(interp, argv[1], &newflag) != TCL_OK) {
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int oldflag = static_cast<int>(director->CheckRestartCrc());
  director->SetRestartCrcCheck(static_cast<BOOL>(newflag));

  char buf[64];
  Oc_Snprintf(buf,sizeof(buf),"%d",oldflag);
  return String(buf);
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_GetMifCrc --
 *      Returns the CRC of the buffer representation of the most recent
 *      file read by the current Oxs_Mif object.  It is the value used
 *      by the most recent problem load event (i.e., the Oxs_Mif ReadMif
 *      command).  Returns an error if no problem is loaded.
 *
 * Results:
 *      CRC value.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------------------------- */
String Oxs_GetMifCrc(Oxs_Director* director,Tcl_Interp *interp,
		     int argc,CONST84 char** argv)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0],"\"",(char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  UINT4m crc;
  crc = director->GetMifCrc();
  char buf[64];
  Oc_Snprintf(buf,sizeof(buf),"%lu",
	      static_cast<unsigned long>(crc));
  return String(buf);
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_GetMifParameters --
 *      Returns the MIF parameter string held in the current Oxs_Mif
 *      object.  It is the value used by the most recent problem
 *      load event (i.e., the Oxs_Mif ReadMif command).  Returns
 *      an error if no problem is loaded.
 *
 * Results:
 *      MIF parameter string.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_GetMifParameters(Oxs_Director* director,Tcl_Interp *interp,
			    int argc,CONST84 char** argv)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0],"\"",(char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  return director->GetMifParameters();
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_GetStage --
 *      Interactively get the Stage of the problem being solved.
 *
 * Results:
 *      Current stage.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

String Oxs_GetStage(Oxs_Director* director,Tcl_Interp *interp,
		    int argc,CONST84 char** argv)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     "\"", (char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int stage = director->GetStage();
  char buf[64];
  Oc_Snprintf(buf,sizeof(buf),"%d",stage);
  return String(buf);
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_GetRunCounts --
 *      Returns a list consisting of the current iteration, stage, and
 *   number of stages.
 *
 * Results:
 *      String form of the list.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

String Oxs_GetRunStateCounts(Oxs_Director* director,Tcl_Interp *interp,
			     int argc,CONST84 char** argv)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     "\"", (char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int iteration = director->GetIteration();
  int stage = director->GetStage();
  int stage_count = director->GetNumberOfStages();

  char buf[1024];
  Oc_Snprintf(buf,sizeof(buf),"%d %d %d",iteration,stage,stage_count);
  return String(buf);
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_SetStage --
 *      Interactively set the Stage of the problem being solved.
 *
 * Results:
 *      Returns the actual stage.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

String Oxs_SetStage(Oxs_Director* director,Tcl_Interp *interp,
		    int argc,CONST84 char** argv)
{
  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " stage_number\"", (char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int requestedStage;
  if (Tcl_GetInt(interp, argv[1], &requestedStage) != TCL_OK) {
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int stage = director->SetStage(requestedStage);

  char buf[64];
  Oc_Snprintf(buf,sizeof(buf),"%d",stage);
  return String(buf);
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_GetMif --
 *      Interactively get the MIF handle of the problem being solved.
 *
 * Results:
 *      MIF handle as a string.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

String Oxs_GetMif(Oxs_Director* director,Tcl_Interp *interp,
		  int argc, CONST84 char** argv)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     "\"", (char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }
  return director->GetMifHandle();
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ProbInit --
 *      Micromagnetic problem initialization
 *
 * Results:
 *      None (an empty string is returned)
 *
 * Side effects:
 *      Creates a slave interpreter
 *
 *----------------------------------------------------------------------
 */

String Oxs_ProbInit(Oxs_Director* director,Tcl_Interp *interp,
		    int argc,CONST84 char** argv)
{
  if (argc != 3) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0]," filename mif_parameters\"",
		     (char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int result = director->ProbInit(argv[1],argv[2]);
  if(result!=TCL_OK) {
    throw OxsCmdsProcTclException(result);
  }

  return String("");
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ProbReset --
 *      Resets micromagnetic problem
 *
 * Results:
 *      None (an empty string is returned)
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_ProbReset(Oxs_Director* director,Tcl_Interp *interp,
		     int argc,CONST84 char** argv)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0],"\"",(char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  int restart_flag = director->SetRestartFlag(0); // Disable restart
  /// flag during reset operation
  try {
    int result = director->ProbReset();
    director->SetRestartFlag(restart_flag);
    if(result!=TCL_OK) {
      throw OxsCmdsProcTclException(result);
    }
  } catch (...) {
    director->SetRestartFlag(restart_flag);
    throw;
  }

  return String("");
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_Run --
 *      Runs micromagnetic problem.
 *
 * Results:
 *      String containing list of events.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_Run(Oxs_Director* director,Tcl_Interp *interp,
	    int argc,CONST84 char** argv)
{ // On success, returns a list of events with associated state id
  // info.  Each item in the list has the form
  //      <event_type state_id stage_number iteration_count>
  // In the future, we might want to change this to return instead
  // some general state reference that the script can query for
  // itself to pull out the info that it needs.
  // NOTE: Each element of the "results" vector holds a read lock
  //  on the associated state, until such time as "results" is
  //  destroyed.
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be \"",
		     argv[0],"\"",(char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  vector<OxsRunEvent> results; // Note: Each OxsRunEvent includes
  /// a key tied to a simulation state.  That key will hold a read
  /// lock on the state until the OxsRunEvent is destroyed.  The
  /// easiest way to insure proper destruction, that works even
  /// across exceptions, is for results to be an automatic (local)
  /// variable.
  try {
    director->ProbRun(results);
  } catch (Oxs_Ext::Error& err) {
    String driver_name;
    if(director->GetDriverInstanceName(driver_name)!=0) {
      OXS_THROW(Oxs_IncompleteInitialization,"no driver identified");
    }
    String head =
      String("Oxs_Ext error inside Run() function of driver ")
      + driver_name + String(" --- ");
    err.Prepend(head);
    throw;
  } catch (Oc_Exception& err) {
    String driver_name;
    if(director->GetDriverInstanceName(driver_name)!=0) {
      OXS_THROW(Oxs_IncompleteInitialization,"no driver identified");
    }
    String head =
      String("Error inside Run() function of driver ")
	+ driver_name
	+ String(" --- ");
    err.Prepend(head);
    throw;
  }

  Tcl_ResetResult(interp); // Safety: Wipe away any accidental
  /// scribblings in the result area.

  String whole_list;
  if(!results.empty()) {
    // Build event list
    vector<String> event_list;
    vector<OxsRunEvent>::const_iterator it;
    for(it=results.begin();it!=results.end();++it) {
      const Oxs_SimState* state = it->state_key.GetPtr();
      if(state==NULL || state->Id()==0) {
	OXS_THROW(Oxs_ProgramLogicError,"Invalid state returned");
      }
      static char event_buf[512];
      switch(it->event) {
      case OXS_STEP_EVENT:
	Oc_Snprintf(event_buf,sizeof(event_buf),
		    "STEP %u %u %u",
		    state->Id(),
		    state->stage_number,
		    state->iteration_count);
	break;
      case OXS_STAGE_DONE_EVENT:
	Oc_Snprintf(event_buf,sizeof(event_buf),
		    "STAGE_DONE %u %u %u",
		    state->Id(),
		    state->stage_number,
		    state->iteration_count);
	break;
      case OXS_RUN_DONE_EVENT:
	Oc_Snprintf(event_buf,sizeof(event_buf),
		    "RUN_DONE %u %u %u",
		    state->Id(),
		    state->stage_number,
		    state->iteration_count);
	break;
      case OXS_INVALID_EVENT:
        OXS_THROW(Oxs_ProgramLogicError,"Invalid event detected");
        break;
      default:
        {
          char buf[1024];
          Oc_Snprintf(buf,sizeof(buf),
                      "Unknown event type returned: %d",
                      static_cast<int>(it->event));
          OXS_THROW(Oxs_ProgramLogicError,buf);
        }
      }
      event_list.push_back(String(event_buf));
    }
    whole_list = Oxs_MergeList(event_list);
  }

  return whole_list;
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_DriverProperties --
 *      Returns properties of the driver.
 *
 * Results:
 *      Driver properties as a string.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_DriverProperties(Oxs_Director* director,Tcl_Interp *interp,
                            int argc,CONST84 char** argv)
{
  BOOL bad_param_count=0;
  BOOL bad_driver_init=0;
  String result;

  if (argc < 2) {
    bad_param_count=1;
  } else if(strcmp("name",argv[1])==0) {
    if(argc!=2) {
      bad_param_count=1;
    } else {
      String name;
      if(director->GetDriverInstanceName(name)!=0) {
	bad_driver_init = 1;
      } else {
	result = name;
      }
    }
  } else if(strcmp("attributes",argv[1])==0) {
    if(argc!=2) {
      bad_param_count=1;
    } else {
      vector<String> labels;
      if(director->GetDriverAttributes(labels) != 0) {
	bad_driver_init = 1;
      } else {
	result = Oxs_MergeList(labels);
      }
    }
  } else if(strcmp("value",argv[1])==0) {
    if(argc!=3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
		       argv[0]," value <label>\"",(char *) NULL);
      throw OxsCmdsProcTclException(TCL_ERROR);
    } else {
      String value;
      int errcode
	= director->GetDriverAttributeValue(String(argv[2]),value);
      if(errcode==0) {
	result = value;
      } else if(errcode==1) {
	bad_driver_init = 1;
      } else {
	Tcl_AppendResult(interp, "Label \"", argv[2],
			 "\" not recognized by driver.",
			 (char *) NULL);
	throw OxsCmdsProcTclException(TCL_ERROR);
      }
    }
  } else {
    Tcl_AppendResult(interp, "bad option \"",argv[1],
		     "\": must be name, attributes or value",
		     (char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  if(bad_param_count) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0]," option ?arg?\"",(char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  if(bad_driver_init) {
    Tcl_AppendResult(interp, "Driver not initialized.",
		     (char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }

  return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListRegisteredExtClasses --
 *      Lists Ext child classes that are loaded and registered
 *
 * Results:
 *      String consisting of Ext child class list.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_ListRegisteredExtClasses(Oxs_Director*,Tcl_Interp *interp,
				    int argc,CONST84 char** argv)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0],"\"",(char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }
  return Oxs_Ext::ListRegisteredChildClasses();
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListExtObjects --
 *      Lists Ext objects that have been created
 *
 * Results:
 *      String.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_ListExtObjects(Oxs_Director* director,Tcl_Interp *interp,
			  int argc,CONST84 char** argv)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0],"\"",(char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }
  return director->ListExtObjects();
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListEnergyObjects --
 *      Lists Energy objects that have been created
 *
 * Results:
 *      String.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_ListEnergyObjects(Oxs_Director* director,Tcl_Interp *interp,
			     int argc,CONST84 char** argv)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0],"\"",(char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }
  return director->ListEnergyObjects();
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListOutputObjects --
 *      Lists all output objects that are registered with the director.
 *
 * Results:
 *      String.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_ListOutputObjects(Oxs_Director* director, Tcl_Interp *interp,
			     int argc, CONST84 char** argv)
{
    if (argc != 1) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0], "\"", (char *) NULL);
	throw OxsCmdsProcTclException(TCL_ERROR);
    }

    vector<String> outputs;
    director->ListOutputObjects(outputs);

    return Oxs_MergeList(outputs);
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_OutputGet--
 *      Calls output method of director.
 *
 * Results:
 *      Output is written into *interp's result.  Return value is
 *      an empty string.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
String Oxs_OutputGet(Oxs_Director* director,Tcl_Interp *interp,
		     int argc,CONST84 char** argv)
{
    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0], " output ...\"", (char *) NULL);
	throw OxsCmdsProcTclException(TCL_ERROR);
    }
    int errcode = director->Output(argv[1], interp, argc-2, argv+2);
    if(errcode!=TCL_OK) {
	throw OxsCmdsProcTclException(errcode);
    }
    return String("");
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_OutputNames --
 *      Retrieves names that describe an output.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      Sets result of interp to a triple of strings giving the
 *      source, name, and type of output.
 *
 *----------------------------------------------------------------------
 */
String Oxs_OutputNames(Oxs_Director* director, Tcl_Interp *interp,
		       int argc, CONST84 char** argv)
{
    if (argc != 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0], " output\"", (char *) NULL);
	throw OxsCmdsProcTclException(TCL_ERROR);
    }
    vector<String> names;
    director->OutputNames(argv[1],names);
    return Oxs_MergeList(names);
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_DirectorDevelopTest --
 *      Hook for development/debug testing.
 *
 * Results:
 *      Results string.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
Oxs_CmdProc	Oxs_DirectorDevelopTest;
String Oxs_DirectorDevelopTest(Oxs_Director* director,Tcl_Interp *interp,
			       int argc,CONST84 char** argv)
{
  if(argc!=2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
		     argv[0]," count\"",(char *) NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }
  String out;
  int result_code = director->DoDevelopTest(String(argv[1]),out);
  if(result_code!=0) {
    char buf[256];
    Oc_Snprintf(buf,sizeof(buf),"%d",result_code);
    Tcl_AppendResult(interp,
		     "Test error; error code: ",buf,
		     "\nSee DoDevelopTest source code in"
		     " oommf/app/oxs/base/director.cc"
		     " for details.",
		     (char *)NULL);
    throw OxsCmdsProcTclException(TCL_ERROR);
  }
  return out;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ProbRelease --
 *      Releases micromagnetic problem statement.  Depending on OS,
 *      this may free some memory, but is intended mainly to put
 *      the solver in a well-defined state.
 *  NB: This routine is unique in that it is called directly by
 *      the interpreter, rather than being routed through the
 *      switchboard routine.  This is to protect against potential
 *      reentrancy problems, and is justified by the special status
 *      of this routine as part of the error recovery path.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int Oxs_ProbRelease(ClientData cd,Tcl_Interp *interp,
		   int argc,CONST84 char**)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be just",
		     " \"Oxs_ProbRelease\"",
		     (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;

  int result = TCL_ERROR;
  try {
    director->Release();
    result = TCL_OK;
  } catch (Oxs_Ext::Error& err) {
    Tcl_AppendResult(interp,"Error thrown from inside"
		     " director ProbRelease function---",
		     err.c_str(),(char *) NULL);
    result = TCL_ERROR; // Safety
  } catch (Oc_Exception& err) {
    Tcl_AppendResult(interp,"Error thrown from inside"
		     " director ProbRelease function---",
		     err.MessageType(),"---",
		     err.MessageText().c_str(),
		     (char *) NULL);
    result = TCL_ERROR; // Safety
  } catch (EXCEPTION& err) {
    Tcl_AppendResult(interp,"Library error thrown from inside"
		     " director ProbRelease function---",
		     err.what(),(char *) NULL);
    result = TCL_ERROR; // Safety
#ifndef NDEBUG
  } catch (...) {
    Tcl_AppendResult(interp,"Unrecognized error thrown from inside"
		     " director ProbRelease function.",
		     (char *) NULL);
    result = TCL_ERROR; // Safety
#endif // NDEBUG
  }

  return result;
}



/*
 *----------------------------------------------------------------------
 *
 * OxsRegisterInterfaceCommands --
 *      Registers each of the above routines with the Tcl interpreter
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
void OxsRegisterInterfaceCommands(Oxs_Director* director,
				  Tcl_Interp *interp)
{
#define REGCMD(foocmd,fooname) \
  Tcl_CreateCommand(interp,Oc_AutoBuf(#foocmd),OxsCmdsSwitchboard,   \
	    new OxsCmdsClientData(director,foocmd,String(fooname)),  \
	    OxsCmdsCleanup)
  REGCMD(Oxs_SetRestartFlag,"director SetRestartFlag");
  REGCMD(Oxs_SetRestartCrcCheck,"director SetRestartCrcCheck");
  REGCMD(Oxs_GetMifCrc,"director GetMifCrc");
  REGCMD(Oxs_GetMifParameters,"director GetMifParameters");
  REGCMD(Oxs_GetStage,"director GetStage");
  REGCMD(Oxs_GetRunStateCounts,"Oxs_GetRunStateCounts");
  REGCMD(Oxs_SetStage,"director SetStage");
  REGCMD(Oxs_GetMif,"director GetMifHandle");
  REGCMD(Oxs_ProbInit,"director ProbInit");
  REGCMD(Oxs_ProbReset,"director ProbReset");
  REGCMD(Oxs_Run,"Oxs_Run");
  REGCMD(Oxs_DriverProperties,"Oxs_DriverProperties");
  REGCMD(Oxs_ListRegisteredExtClasses,"Oxs_ListRegisteredExtClasses");
  REGCMD(Oxs_ListExtObjects,"director ListExtObjects");
  REGCMD(Oxs_ListEnergyObjects,"director ListEnergyObjects");
  REGCMD(Oxs_ListOutputObjects,"director ListOutputObjects");
  REGCMD(Oxs_OutputGet,"director Output");
  REGCMD(Oxs_OutputNames,"director OutputNames");
  REGCMD(Oxs_DirectorDevelopTest,"Oxs_DirectorDevelopTest");
#undef REGCMD
  // The ProbRelease routine is special
  Tcl_CreateCommand(interp,Oc_AutoBuf("Oxs_ProbRelease"),
                    Oxs_ProbRelease,director,NULL);
}
